import os

from berglas import VERSION


CONTINUE_ON_ERROR = os.environ.get("BERGLAS_CONTINUE_ON_ERROR", "false").lower() == "true"
REFERENCE_PREFIX = "berglas://"
USER_AGENT = f"berglas/{VERSION} (https://github.com/maroux/berglas)"
METADATA_KMS_KEY = "berglas-kms-key"



import abc
import http.client
import os
from pathlib import Path
from typing import Dict

import google.auth
from google_auth_httplib2 import AuthorizedHttp
from googleapiclient import discovery
from googleapiclient.errors import HttpError
from googleapiclient.http import set_user_agent

from berglas.constants import USER_AGENT
from berglas.exceptions import AutoException


class RuntimeEnv(abc.ABC):
    def _build(self, service_name: str, version: str):
        resource = discovery.build(service_name, version, cache_discovery=False)
        set_user_agent(resource._http, USER_AGENT)
        return resource

    def get_env(self) -> Dict[str, str]:
        raise NotImplementedError

    @staticmethod
    def _get_from_metadata(path: str) -> str:
        path = f"http://metadata.google.internal/computeMetadata/v1/{path}"

        http_client = discovery.build_http()
        set_user_agent(http_client, USER_AGENT)
        response, content = http_client.request(path, headers={"Metadata-Flavor": "Google", "User-Agent": USER_AGENT})
        if response.status != http.client.OK.value:
            raise AutoException(f"failed to get metadata for: {path}: {content}")

        return content.decode()


class CloudFunctionEnv(RuntimeEnv):
    def get_env(self) -> Dict[str, str]:
        name = (
            f"projects/{os.environ['X_GOOGLE_GCP_PROJECT']}/locations/{os.environ['X_GOOGLE_FUNCTION_REGION']}/"
            f"functions/{os.environ['X_GOOGLE_FUNCTION_NAME']}"
        )

        cloudfunctions_v1 = self._build("cloudfunctions", "v1")
        try:
            function = cloudfunctions_v1.projects().locations().functions().get(name=name).execute()
        except HttpError as e:
            if e.resp.status == http.client.NOT_FOUND.value:
                raise AutoException("function not found")
            elif e.resp.status == http.client.UNAUTHORIZED.value:
                raise AutoException("permission denied to fetch function info")

            raise AutoException(f"unable to fetch function: {e}")

        return function["environmentVariables"]


class CloudRunEnv(RuntimeEnv):
    def get_env(self) -> Dict[str, str]:
        revision = os.environ["K_REVISION"]

        project = self._get_from_metadata("project/project-id")
        zone = self._get_from_metadata("instance/zone")
        region = Path(zone).name[:-2]

        name = f"projects/{project}/locations/{region}/revisions/{revision}"

        cloudrun = self._build("run", "v1alpha1")
        try:
            service = cloudrun.projects().locations().revisions().get(name=name).execute()
        except HttpError as e:
            if e.resp.status == http.client.NOT_FOUND.value:
                raise AutoException("cloud run service not found")
            elif e.resp.status == http.client.UNAUTHORIZED.value:
                raise AutoException("permission denied to fetch cloud run info")

            raise AutoException(f"unable to fetch cloud run service: {e.content}")

        if "spec" not in service or "containers" not in service["spec"] or not service["spec"]["containers"]:
            raise AutoException(f"unable to get env from cloud run")

        env = service["spec"]["containers"][0]["env"]
        return {v["name"]: v["value"] for v in env}


class AppEngineEnv(RuntimeEnv):
    def get_env(self) -> Dict[str, str]:
        version = os.environ["GAE_VERSION"]
        service = os.environ["GAE_SERVICE"]

        project = self._get_from_metadata("project/project-id")

        gae = self._build("appengine", "v1")
        try:
            service_info = (
                gae.apps()
                .services()
                .versions()
                .get(appsId=project, servicesId=service, versionsId=version, view="FULL")
                .execute()
            )
        except HttpError as e:
            if e.resp.status == http.client.NOT_FOUND.value:
                raise AutoException("app engine service not found")
            elif e.resp.status == http.client.UNAUTHORIZED.value:
                raise AutoException("permission denied to fetch app engine info")

            raise AutoException(f"unable to fetch app engine service: {e.content}")

        return service_info["envVariables"]


def detect_runtime_environment() -> RuntimeEnv:
    """
    Finds the most likely runtime environment
    """
    if os.environ.get("X_GOOGLE_FUNCTION_NAME"):
        return CloudFunctionEnv()

    if os.environ.get("K_REVISION"):
        return CloudRunEnv()

    if os.environ.get("GAE_SERVICE"):
        return AppEngineEnv()

    raise AutoException("unknown runtime")



import base64
import binascii
import os
import stat
import tempfile
from urllib.parse import parse_qs, urlparse

from Crypto.Cipher import AES
from google.api_core.client_info import ClientInfo as APICoreClientInfo
from google.api_core.exceptions import GoogleAPICallError, RetryError
from google.api_core.gapic_v1.client_info import ClientInfo as GAPIClientInfo
from google.cloud import storage, kms

from berglas.constants import METADATA_KMS_KEY, REFERENCE_PREFIX, USER_AGENT
from berglas.exceptions import AutoException


def is_reference(key: str) -> bool:
    """
    Does this key represent a berglas reference
    """
    return key.startswith(REFERENCE_PREFIX)


def _decrypt_gcm(dek: bytes, ciphertext: bytes) -> bytes:
    # TODO use some constant from crypto library instead of hardcoding here
    nonce_size = 12
    tag_size = 16
    if len(ciphertext) < nonce_size + tag_size:
        raise AutoException("malformed ciphertext")
    nonce, ciphertext, tag = ciphertext[:nonce_size], ciphertext[nonce_size:-tag_size], ciphertext[-tag_size:]
    cipher = AES.new(dek, AES.MODE_GCM, nonce=nonce)
    try:
        return cipher.decrypt_and_verify(ciphertext, tag)
    except ValueError:
        raise AutoException("failed to decrypt ciphertext")


class Client:
    storage_client: storage.Client
    kms_client: kms.KeyManagementServiceClient

    def __init__(self):
        self.storage_client = storage.Client(client_info=APICoreClientInfo(user_agent=USER_AGENT))
        self.kms_client = kms.KeyManagementServiceClient(client_info=GAPIClientInfo(user_agent=USER_AGENT))

    def _access(self, bucket_name: str, path: str) -> bytes:
        """
        Get plaintext value of the secret stored at path in bucket
        """
        blob = self.storage_client.bucket(bucket_name).get_blob(path, client=self.storage_client)
        if not blob:
            raise AutoException("secret object not found")

        if not blob.metadata or not blob.metadata.get(METADATA_KMS_KEY):
            raise AutoException("missing kms key in secret metadata")

        key = blob.metadata[METADATA_KMS_KEY]
        data = blob.download_as_string()
        parts = data.split(b":", maxsplit=1)
        if len(parts) < 2:
            raise AutoException("invalid ciphertext: not enough parts")

        try:
            enc_dek = base64.decodebytes(parts[0])
        except binascii.Error:
            raise AutoException("invalid ciphertext: failed to parse dek")

        try:
            ciphertext = base64.decodebytes(parts[1])
        except binascii.Error:
            raise AutoException("invalid ciphertext: failed to parse ciphertext")

        try:
            response = self.kms_client.decrypt(key, enc_dek, path.encode("UTF8"))
            dek = response.plaintext
        except (GoogleAPICallError, RetryError, ValueError):
            raise AutoException("failed to decrypt dek")

        return _decrypt_gcm(dek, ciphertext)

    def resolve(self, value: str) -> bytes:
        """
        Resolves a berglas reference value to the plain text secret
        """
        if not is_reference(value):
            raise AutoException("not a berglas reference")

        parsed = urlparse(value)
        bucket = parsed.netloc
        obj = parsed.path.lstrip("/")
        destination = parse_qs(parsed.query).get("destination", [""])[0]
        tmpfile = None

        if destination in ["tmpfile", "tempfile"]:
            tmpfile = tempfile.NamedTemporaryFile(prefix="berglas-", delete=False)
        elif destination:
            tmpfile = open(destination, "wb")

        plaintext = self._access(bucket, obj)

        if tmpfile:
            os.chmod(tmpfile.name, stat.S_IRUSR | stat.S_IWUSR)
            tmpfile.write(plaintext)
            plaintext = tmpfile.name.encode()
            tmpfile.close()

        return plaintext

    def replace(self, key: str, value: str = None) -> None:
        """
        Replace a berglas reference represented by value and update key in environment with plain text value
        :param value: if specified use this value instead of reading from os.environ
        """
        plaintext = self.resolve(value or os.environ[key])
        # don't assume UTF-8, or system encoding even
        os.environb[key.encode()] = plaintext



import logging
from typing import Dict

from berglas.constants import CONTINUE_ON_ERROR
from berglas.exceptions import AutoException
from berglas.resolver import Client, is_reference
from berglas.runtime import RuntimeEnv, detect_runtime_environment


def _handle_error(e: BaseException) -> None:
    logging.exception(e.args[0])
    if not CONTINUE_ON_ERROR:
        raise


def _filter_refs(runtime_env: RuntimeEnv) -> Dict[str, str]:
    env_vars = runtime_env.get_env()
    return {k: v for k, v in env_vars.items() if is_reference(v)}


def auto() -> None:
    """
    Automatically resolve and update environment with plain text secrets
    """
    logging.debug("automatically resolving all berglas references from environment")
    try:
        runtime_env = detect_runtime_environment()

        env_var_refs = _filter_refs(runtime_env)
        if not env_var_refs:
            logging.warning("berglas auto was included, but no secrets were found in the environment")
            return

        client = Client()
        for k, v in env_var_refs.items():
            client.replace(k, v)
    except AutoException as e:
        _handle_error(e)
    except BaseException as e:
        _handle_error(e)


if __name__ == "berglas.auto":
    auto()



"""A setuptools based setup module.
See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

import ast
from codecs import open
from os import path
import re
from setuptools import setup, find_packages


cwd = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(cwd, "README.rst"), encoding="utf-8") as f:
    long_description = f.read()

_version_re = re.compile(r"VERSION\s+=\s+(.*)")

with open("berglas/__init__.py", "rb") as f:
    version = str(ast.literal_eval(_version_re.search(f.read().decode("utf-8")).group(1)))

tests_require = ["pytest", "flake8", "mypy", "ipdb", "coverage", "coveralls", "pytest-cov", "black"]

setup(
    name="berglas",
    version=version,
    description="Berglas Python Library",
    long_description=long_description,
    url="https://github.com/maroux/berglas-python",
    author="Aniruddha Maru",
    license="Apache Software License (Apache License 2.0)",
    maintainer="Aniruddha Maru",
    maintainer_email="aniruddhamaru@gmail.com",
    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        "Natural Language :: English",
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: System Administrators",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: Apache Software License",
    ],
    python_requires=">=3.6",
    keywords="secrets gcs gcp",
    # https://mypy.readthedocs.io/en/latest/installed_packages.html
    package_data={"berglas": ["py.typed"]},
    packages=find_packages(exclude=["tests", "tests.*"]),
    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=["pycryptodome", "google-api-python-client", "google-cloud-storage", "google-cloud-kms"],
    tests_require=tests_require,
    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    extras_require={"dev": ["flake8"], "test": tests_require, "publish": ["twine", "wheel"]},
    include_package_data=True,
)
